using System;
using System.Collections.Generic;
using System.Collections;
using UnityEngine;
using System.Linq;
using TMPro;
using UnityEngine.UI;
using HarmonyLib;
using Gizmo;
using HMLLibrary;

public class WorldEditor : Mod
{
    public static AssetBundle assetbundle;
    public static WorldEditor instance;

    GameObject CurrentItemCanvas;

    // Block Editor Variables
    public static TMP_InputField xpos;
    public static TMP_InputField ypos;
    public static TMP_InputField zpos;
    public static TMP_InputField xrot;
    public static TMP_InputField yrot;
    public static TMP_InputField zrot;
    public static CanvasGroup BlockEditorCanvas;
    public static Item_Base BlockEditorItem;
    public static GameObject[] MeshSelectors = new GameObject[20];
    public static Material GreenMaterial;
    public static Block selectedblock;
    public static TextMeshProUGUI blockname;
    public static bool isBlockEditorOpened;

    public static Shader LineShader;

    // Prefab Spawner Variables
    public static Item_Base PrefabSpawnerItem;
    public static CanvasGroup PrefabSpawnerCanvas;
    public static bool isPrefabSpawnerOpened;
    public static PrefabSpawnerPoint PrefabSpawnerPoint;

    public IEnumerator Start()
    {
        instance = this;
        AssetBundleCreateRequest request = AssetBundle.LoadFromMemoryAsync(GetEmbeddedFileBytes("worldeditor.assets"));
        yield return request;
        assetbundle = request.assetBundle;

        LineShader = assetbundle.LoadAsset<Material>("Line").shader;

        CurrentItemCanvas = Instantiate(assetbundle.LoadAsset<GameObject>("CurrentItemCanvas"), transform);
        CurrentItemCanvas.SetActive(false);

        InitializeBlockEditor();
        InitializePrefabSpawner();

        Debug.Log("<color=green><b>WorldEditor has been loaded!</b></color>");
    }
    public void InitializeBlockEditor()
    {
        for (int i = 0; i < 20; i++) { MeshSelectors[i] = Instantiate(assetbundle.LoadAsset<GameObject>("MeshSelector"), transform); }
        MeshSelectors.ToList().ForEach((ms) => ms.SetActive(false));
        GreenMaterial = assetbundle.LoadAsset<Material>("GreenMaterial");
        BlockEditorItem = assetbundle.LoadAsset<Item_Base>("BlockEditorTool");
        RAPI.RegisterItem(BlockEditorItem);

        GameObject temp = Instantiate(assetbundle.LoadAsset<GameObject>("BlockEditorMenu"), transform);
        BlockEditorCanvas = temp.GetComponent<CanvasGroup>();

        blockname = BlockEditorCanvas.transform.Find("Background").Find("BlockName").GetComponent<TextMeshProUGUI>();
        xpos = BlockEditorCanvas.transform.Find("Background").Find("TransformPosition").Find("XPOS").GetComponent<TMP_InputField>();
        ypos = BlockEditorCanvas.transform.Find("Background").Find("TransformPosition").Find("YPOS").GetComponent<TMP_InputField>();
        zpos = BlockEditorCanvas.transform.Find("Background").Find("TransformPosition").Find("ZPOS").GetComponent<TMP_InputField>();
        xrot = BlockEditorCanvas.transform.Find("Background").Find("TransformRotation").Find("XROT").GetComponent<TMP_InputField>();
        yrot = BlockEditorCanvas.transform.Find("Background").Find("TransformRotation").Find("YROT").GetComponent<TMP_InputField>();
        zrot = BlockEditorCanvas.transform.Find("Background").Find("TransformRotation").Find("ZROT").GetComponent<TMP_InputField>();
        BlockEditorCanvas.transform.Find("Background").Find("DestroyButton").GetComponent<Button>().onClick.AddListener(DestroySelectedBlock);
        BlockEditorCanvas.transform.Find("Background").Find("DuplicateButton").GetComponent<Button>().onClick.AddListener(DuplicateSelectedBlock);
        BlockEditorCanvas.transform.Find("Background").Find("ApplyButton").GetComponent<Button>().onClick.AddListener(ApplyBlockEditorChanges);
        BlockEditorCanvas.transform.Find("Background").Find("Topbar").Find("CloseButton").GetComponent<Button>().onClick.AddListener(() => { SetBlockEditorMenuState(false); });

        BlockEditorCanvas.transform.Find("Background").Find("GizmoButtons").Find("MoveTransform").GetComponent<Button>().onClick.AddListener(() =>
        {
            Camera.main.gameObject.GetComponent<TransformGizmo>().SetMoveType();
        });

        BlockEditorCanvas.transform.Find("Background").Find("GizmoButtons").Find("RotateTransform").GetComponent<Button>().onClick.AddListener(() =>
        {
            Camera.main.gameObject.GetComponent<TransformGizmo>().SetRotateType();
            Debug.Log("Type changed to" + Camera.main.gameObject.GetComponent<TransformGizmo>().transformType);
        });

        BlockEditorCanvas.transform.Find("Background").Find("GizmoButtons").Find("ChangeSpace").GetComponent<Button>().onClick.AddListener(() =>
        {
            Camera.main.gameObject.GetComponent<TransformGizmo>().ToggleSpace();
        });

        BlockEditorCanvas.alpha = 0;
        BlockEditorCanvas.interactable = false;
        BlockEditorCanvas.blocksRaycasts = false;

        GameObject WandItem = assetbundle.LoadAsset<GameObject>("BlockEditorWandObject");
        WandItem.AddComponent<BlockEditorWand>();
        RAPI.SetItemObject(BlockEditorItem, WandItem);
    }
    public void SetBlockEditorMenuState(bool value)
    {
        RAPI.ToggleCursor(value);
        BlockEditorCanvas.alpha = value ? 1 : 0;
        BlockEditorCanvas.interactable = value ? true : false;
        BlockEditorCanvas.blocksRaycasts = value ? true : false;
        isBlockEditorOpened = value;
        if (!value)
        {
            selectedblock = null;
            MeshSelectors.ToList().ForEach((ms) => ms.SetActive(false));
        }
        if (!value)
        {
            FindObjectsOfType<TransformGizmo>().ToList().ForEach((b) => Destroy(b));
        }
    }
    public void InitializePrefabSpawner()
    {
        PrefabSpawnerItem = assetbundle.LoadAsset<Item_Base>("PrefabSpawnerTool");
        RAPI.RegisterItem(PrefabSpawnerItem);

        GameObject temp = Instantiate(assetbundle.LoadAsset<GameObject>("PrefabSpawnerMenu"), transform);
        PrefabSpawnerCanvas = temp.GetComponent<CanvasGroup>();
        PrefabSpawnerCanvas.transform.Find("Background").Find("Topbar").Find("CloseButton").GetComponent<Button>().onClick.AddListener(() => { SetPrefabSpawnerMenuState(false); });
        PrefabSpawnerCanvas.transform.Find("Background").Find("Scroll View").Find("Viewport").Find("Content").Find("BirchTree").GetComponentInChildren<Button>().onClick.AddListener(() => PrefabSpawnerMethods.SpawnCropplot(ItemManager.GetItemByIndex(284)));
        PrefabSpawnerCanvas.transform.Find("Background").Find("Scroll View").Find("Viewport").Find("Content").Find("MangoTree").GetComponentInChildren<Button>().onClick.AddListener(() => PrefabSpawnerMethods.SpawnCropplot(ItemManager.GetItemByIndex(167)));
        PrefabSpawnerCanvas.transform.Find("Background").Find("Scroll View").Find("Viewport").Find("Content").Find("PalmTree").GetComponentInChildren<Button>().onClick.AddListener(() => PrefabSpawnerMethods.SpawnCropplot(ItemManager.GetItemByIndex(31)));
        PrefabSpawnerCanvas.transform.Find("Background").Find("Scroll View").Find("Viewport").Find("Content").Find("PineTree").GetComponentInChildren<Button>().onClick.AddListener(() => PrefabSpawnerMethods.SpawnCropplot(ItemManager.GetItemByIndex(283)));
        PrefabSpawnerCanvas.alpha = 0;
        PrefabSpawnerCanvas.interactable = false;
        PrefabSpawnerCanvas.blocksRaycasts = false;

        GameObject WandItem = assetbundle.LoadAsset<GameObject>("PrefabSpawnerWandObject");
        WandItem.AddComponent<PrefabSpawnerWand>();
        RAPI.SetItemObject(PrefabSpawnerItem, WandItem);
    }
    public void SetPrefabSpawnerMenuState(bool value)
    {
        RAPI.ToggleCursor(value);
        PrefabSpawnerCanvas.alpha = value ? 1 : 0;
        PrefabSpawnerCanvas.interactable = value ? true : false;
        PrefabSpawnerCanvas.blocksRaycasts = value ? true : false;
        isPrefabSpawnerOpened = value;
        if (!value)
        {
            PrefabSpawnerPoint = null;
        }
    }
    public void DestroySelectedBlock()
    {
        if (selectedblock != null)
        {
            BlockCreator.RemoveBlockNetwork(selectedblock, RAPI.GetLocalPlayer(), true);
            SetBlockEditorMenuState(false);
        }
    }
    public void DuplicateSelectedBlock()
    {
        if (selectedblock != null)
        {
            GameObject block = Instantiate(selectedblock.gameObject, selectedblock.transform.position, selectedblock.transform.rotation, selectedblock.transform.parent);
            block.transform.localScale = selectedblock.transform.localScale;
            SO_ColorValue[] basecolors = Resources.LoadAll<SO_ColorValue>("colors/");
            Block fake = block.GetComponent<Block>();
            Block original = selectedblock;
            fake.SetInstanceColorAndPattern(original.currentColorA, original.currentColorB, 0, original.currentPatternIndex2);
            FindObjectOfType<HNotify>().AddNotification(HNotify.NotificationType.normal, "Successfully Duplicated Block!", 3, HNotify.CheckSprite);

            if (block.GetComponent<Cropplot>())
            {
                int i = 0;
                block.GetComponent<Cropplot>().GetSlots().ForEach((slot) => { block.GetComponent<Cropplot>().SetPlantComplete(i); i++; });
            }
            block.GetComponent<Block>().OnFinishedPlacement();
        }
    }
    public void ApplyBlockEditorChanges()
    {
        if (selectedblock != null)
        {
            try
            {
                float _xpos = float.Parse(xpos.text.Replace(".", ","));
                float _ypos = float.Parse(ypos.text.Replace(".", ","));
                float _zpos = float.Parse(zpos.text.Replace(".", ","));

                float _xrot = float.Parse(xrot.text.Replace(".", ","));
                float _yrot = float.Parse(yrot.text.Replace(".", ","));
                float _zrot = float.Parse(zrot.text.Replace(".", ","));

                selectedblock.transform.localPosition = new Vector3(_xpos, _ypos, _zpos);
                selectedblock.transform.localRotation = Quaternion.Euler(_xrot, _yrot, _zrot);

                FindObjectOfType<HNotify>().AddNotification(HNotify.NotificationType.normal, "Successfully Applied Changes!", 3, HNotify.CheckSprite);
            }
            catch
            {
                FindObjectOfType<HNotify>().AddNotification(HNotify.NotificationType.scaling, "Invalid Value Specified!", 3, HNotify.ErrorSprite);
            }
        }
    }
    public bool IsHoldingItem(Item_Base item)
    {
        return RAPI.GetLocalPlayer() && RAPI.GetLocalPlayer().Inventory.GetSelectedHotbarItem() != null && RAPI.GetLocalPlayer().Inventory.GetSelectedHotbarItem().UniqueName == item.UniqueName;
    }

    public void Update()
    {
        if (RAPI.IsCurrentSceneMainMenu()) { return; }
        if (IsHoldingItem(BlockEditorItem))
        {
            CurrentItemCanvas.SetActive(true);
            CurrentItemCanvas.GetComponentInChildren<TextMeshProUGUI>().text = @"Be careful!
You are holding the fabulous
<color=green>BLOCK EDITOR</color>";
        }
        else if (IsHoldingItem(PrefabSpawnerItem))
        {
            CurrentItemCanvas.SetActive(true);
            CurrentItemCanvas.GetComponentInChildren<TextMeshProUGUI>().text = @"Be careful!
You are holding the fabulous
<color=green>PREFAB SPAWNER</color>";
        }
        else
        {
            CurrentItemCanvas.SetActive(false);
            CurrentItemCanvas.GetComponentInChildren<TextMeshProUGUI>().text = "";
        }

        if (IsHoldingItem(BlockEditorItem) && selectedblock != null)
        {
            blockname.text = selectedblock.buildableItem.settings_Inventory.DisplayName;
            if (!isBlockEditorOpened)
            {
                FindObjectsOfType<TransformGizmo>().ToList().ForEach((b) => Destroy(b));
                TransformGizmo shit = Camera.main.gameObject.AddComponent<TransformGizmo>();
                WorldEditor.BlockEditorCanvas.transform.Find("Background").Find("GizmoButtons").Find("ChangeSpace").GetComponentInChildren<TextMeshProUGUI>().text = Camera.main.gameObject.AddComponent<TransformGizmo>().space.ToString();
                shit.ClearAndAddTarget(selectedblock.transform);
                xpos.text = Math.Round(selectedblock.transform.localPosition.x, 4).ToString().Replace(",", ".");
                ypos.text = Math.Round(selectedblock.transform.localPosition.y, 4).ToString().Replace(",", ".");
                zpos.text = Math.Round(selectedblock.transform.localPosition.z, 4).ToString().Replace(",", ".");

                xrot.text = Math.Round(selectedblock.transform.localRotation.eulerAngles.x, 4).ToString().Replace(",", ".");
                yrot.text = Math.Round(selectedblock.transform.localRotation.eulerAngles.y, 4).ToString().Replace(",", ".");
                zrot.text = Math.Round(selectedblock.transform.localRotation.eulerAngles.z, 4).ToString().Replace(",", ".");
            }
            SetBlockEditorMenuState(true);

            if (Input.GetKeyDown(KeyCode.Escape) && isBlockEditorOpened)
            {
                FindObjectsOfType<TransformGizmo>().ToList().ForEach((b) => Destroy(b));
                SetBlockEditorMenuState(false);
            }
        }
        else
        {
            if (isBlockEditorOpened)
            {
                FindObjectsOfType<TransformGizmo>().ToList().ForEach((b) => Destroy(b));
                SetBlockEditorMenuState(false);
            }
        }

        if (IsHoldingItem(PrefabSpawnerItem) && PrefabSpawnerPoint != null)
        {
            // do my shit here.
            SetPrefabSpawnerMenuState(true);

            if (Input.GetKeyDown(KeyCode.Escape) && isPrefabSpawnerOpened)
            {
                SetPrefabSpawnerMenuState(false);
            }
        }
        else
        {
            if (isPrefabSpawnerOpened)
            {
                SetPrefabSpawnerMenuState(false);
            }
        }
    }

    public void OnModUnload()
    {
        Debug.Log("WorldEditor has been unloaded!");
        Destroy(gameObject);
    }
}

public static class PrefabSpawnerMethods
{
    public static void SpawnCropplot(Item_Base item)
    {
        if (WorldEditor.PrefabSpawnerPoint != null)
        {
            Block b = RAPI.GetLocalPlayer().BlockCreator.CreateBlockCheat(ItemManager.GetItemByIndex(139), WorldEditor.PrefabSpawnerPoint.position, new Vector3(0, 0, 0), DPS.Floor, -1);
            b.transform.localRotation = Quaternion.LookRotation(Vector3.forward, WorldEditor.PrefabSpawnerPoint.normal);
            Cropplot cropplot = b.GetComponent<Cropplot>();
            List<Plant> plants = Traverse.Create(Traverse.Create(cropplot).Field("plantManager").GetValue() as PlantManager).Field("allPlants").GetValue() as List<Plant>;
            Plant plant = plants.Find((p) => p.item.UniqueIndex == item.UniqueIndex);
            if (plant != null)
            {
                cropplot.PlantSeed(plant, SaveAndLoad.GetUniqueObjectIndex(), true);
                cropplot.SetPlantComplete(0);
            }
            b.transform.position = WorldEditor.PrefabSpawnerPoint.position -= (b.transform.up) * 0.2f;
        }
        WorldEditor.instance.SetPrefabSpawnerMenuState(false);
    }
}

public class BlockEditorWand : MonoBehaviour
{
    public void Update()
    {
        if (CanvasHelper.ActiveMenu == MenuType.None && MyInput.GetButtonDown("LMB"))
        {
            RaycastHit hit;
            if (Physics.Raycast(RAPI.GetLocalPlayer().Camera.transform.position, RAPI.GetLocalPlayer().Camera.transform.forward, out hit, 100, LayerMasks.MASK_Block | 1 << LayerMask.NameToLayer("Tree") | 1 << LayerMask.NameToLayer("Item")))
            {
                Block block = hit.transform.GetComponentInParent<Block>();
                if (!block) { block = hit.transform.parent.GetComponentInParent<Block>(); }
                if (block != null)
                {
                    WorldEditor.selectedblock = block;
                }
            }
        }

        if (WorldEditor.selectedblock != null)
        {
            List<MeshFilter> meshFilters = WorldEditor.selectedblock.GetComponentsInChildren<MeshFilter>().ToList().Where((mf) => mf.GetComponent<MeshRenderer>().enabled && mf.gameObject.activeSelf).ToList();
            int id = 0;
            meshFilters.ToList().ForEach((mf) =>
            {
                WorldEditor.MeshSelectors[id].GetComponent<MeshRenderer>().materials = Enumerable.Repeat<Material>(WorldEditor.GreenMaterial, mf.GetComponent<MeshRenderer>().materials.Length).ToArray();
                CreateMeshSelector(mf.mesh, mf.transform, id);
                id++;
            });
            List<SkinnedMeshRenderer> skinnedMeshRenderers = WorldEditor.selectedblock.GetComponentsInChildren<SkinnedMeshRenderer>().ToList().Where((mf) => mf.enabled && mf.gameObject.activeSelf).ToList();
            skinnedMeshRenderers.ToList().ForEach((mr) =>
            {
                WorldEditor.MeshSelectors[id].GetComponent<MeshRenderer>().materials = Enumerable.Repeat<Material>(WorldEditor.GreenMaterial, mr.GetComponent<SkinnedMeshRenderer>().materials.Length).ToArray();
                Mesh mesh = new Mesh(); mr.BakeMesh(mesh);
                CreateMeshSelector(mesh, mr.transform, id);
                id++;
            });
        }
        else
        {
            WorldEditor.MeshSelectors.ToList().ForEach((ms) => ms.SetActive(false));
        }
    }

    public void CreateMeshSelector(Mesh mesh, Transform _transform, int id)
    {
        WorldEditor.MeshSelectors[id].transform.parent = _transform.parent;
        WorldEditor.MeshSelectors[id].GetComponent<MeshFilter>().mesh = mesh;
        WorldEditor.MeshSelectors[id].transform.position = _transform.position;
        WorldEditor.MeshSelectors[id].transform.rotation = _transform.rotation;
        WorldEditor.MeshSelectors[id].transform.localScale = _transform.localScale;
        WorldEditor.MeshSelectors[id].transform.parent = WorldEditor.instance.transform;
        WorldEditor.MeshSelectors[id].SetActive(true);
    }
}

public class PrefabSpawnerWand : MonoBehaviour
{
    public void Update()
    {
        if (CanvasHelper.ActiveMenu == MenuType.None && MyInput.GetButtonDown("LMB"))
        {
            RaycastHit hit;
            if (Physics.Raycast(RAPI.GetLocalPlayer().Camera.transform.position, RAPI.GetLocalPlayer().Camera.transform.forward, out hit, 100, LayerMasks.MASK_Block))
            {
                WorldEditor.PrefabSpawnerPoint = new PrefabSpawnerPoint(hit.point, hit.normal);
            }
        }
    }
}

public class PrefabSpawnerPoint
{
    public Vector3 position;
    public Vector3 normal;

    public PrefabSpawnerPoint(Vector3 _position, Vector3 _normal)
    {
        position = _position;
        normal = _normal;
    }
}

namespace Gizmo
{
    public class TransformGizmo : MonoBehaviour
    {
        public TransformSpace space = TransformSpace.Global;
        public TransformType transformType = TransformType.Move;
        public TransformPivot pivot = TransformPivot.Pivot;
        public KeyCode translationSnapping = KeyCode.LeftControl;
        public Color xColor = new Color(1, 0, 0, 0.8f);
        public Color yColor = new Color(0, 1, 0, 0.8f);
        public Color zColor = new Color(0, 0, 1, 0.8f);
        public Color allColor = new Color(.7f, .7f, .7f, 0.8f);
        public Color selectedColor = new Color(1, 1, 0, 0.8f);
        public Color hoverColor = new Color(1, .75f, 0, 0.8f);
        public float planesOpacity = .5f;
        public float movementSnap = .25f;
        public float rotationSnap = 15f;
        public float handleLength = .25f;
        public float handleWidth = .003f;
        public float planeSize = .035f;
        public float triangleSize = .03f;
        public float boxSize = .03f;
        public int circleDetail = 40;
        public float allMoveHandleLengthMultiplier = 1f;
        public float allRotateHandleLengthMultiplier = 1.4f;
        public float minSelectedDistanceCheck = .01f;
        public float moveSpeedMultiplier = 1f;
        public float rotateSpeedMultiplier = 1f;
        public float allRotateSpeedMultiplier = 20f;
        public bool useFirstSelectedAsMain = true;
        public bool circularRotationMethod;
        public bool forceUpdatePivotPointOnChange = true;
        public bool manuallyHandleGizmo;
        public Action onCheckForSelectedAxis;
        public Action onDrawCustomGizmo;
        public bool isTransforming { get; private set; }
        public Quaternion totalRotationAmount { get; private set; }
        public Axis translatingAxis { get { return nearAxis; } }
        public Axis translatingAxisPlane { get { return planeAxis; } }
        public bool hasTranslatingAxisPlane { get { return translatingAxisPlane != Axis.None && translatingAxisPlane != Axis.Any; } }
        public TransformType transformingType { get { return translatingType; } }

        public Vector3 pivotPoint { get; private set; }
        Vector3 totalCenterPivotPoint;

        public Transform mainTargetRoot { get { return (targetRootsOrdered.Count > 0) ? (useFirstSelectedAsMain) ? targetRootsOrdered[0] : targetRootsOrdered[targetRootsOrdered.Count - 1] : null; } }

        AxisInfo axisInfo;
        Axis nearAxis = Axis.None;
        Axis planeAxis = Axis.None;
        TransformType translatingType;

        AxisVectors handleLines = new AxisVectors();
        AxisVectors handlePlanes = new AxisVectors();
        AxisVectors handleTriangles = new AxisVectors();
        AxisVectors handleSquares = new AxisVectors();
        AxisVectors circlesLines = new AxisVectors();

        //We use a HashSet and a List for targetRoots so that we get fast lookup with the hashset while also keeping track of the order with the list.
        List<Transform> targetRootsOrdered = new List<Transform>();
        Dictionary<Transform, TargetInfo> targetRoots = new Dictionary<Transform, TargetInfo>();
        HashSet<Renderer> highlightedRenderers = new HashSet<Renderer>();
        HashSet<Transform> children = new HashSet<Transform>();

        List<Transform> childrenBuffer = new List<Transform>();
        List<Renderer> renderersBuffer = new List<Renderer>();
        List<Material> materialsBuffer = new List<Material>();

        WaitForEndOfFrame waitForEndOFFrame = new WaitForEndOfFrame();
        Coroutine forceUpdatePivotCoroutine;

        static Material lineMaterial;

        public void SetMoveType()
        {
            transformType = TransformType.Move;
            translatingType = transformType;
        }

        public void SetRotateType()
        {
            transformType = TransformType.Rotate;
            translatingType = transformType;
        }

        public void ToggleSpace()
        {
            space = space == TransformSpace.Global ? TransformSpace.Local : TransformSpace.Global;
            WorldEditor.BlockEditorCanvas.transform.Find("Background").Find("GizmoButtons").Find("ChangeSpace").GetComponentInChildren<TextMeshProUGUI>().text = space.ToString();
        }

        void Awake()
        {
            SetMaterial();
        }

        void OnEnable()
        {
            forceUpdatePivotCoroutine = StartCoroutine(ForceUpdatePivotPointAtEndOfFrame());
        }

        void OnDisable()
        {
            ClearTargets();
            StopCoroutine(forceUpdatePivotCoroutine);
        }

        void Update()
        {
            if (!isTransforming) translatingType = transformType;
            if (manuallyHandleGizmo)
            {
                if (onCheckForSelectedAxis != null) onCheckForSelectedAxis();
            }
            else
            {
                SetNearAxis();
            }

            if (mainTargetRoot == null) return;

            TransformSelected();
        }

        void LateUpdate()
        {
            if (mainTargetRoot == null) return;

            //We run this in lateupdate since coroutines run after update and we want our gizmos to have the updated target transform position after TransformSelected()
            SetAxisInfo();

            if (manuallyHandleGizmo)
            {
                if (onDrawCustomGizmo != null) onDrawCustomGizmo();
            }
            else
            {
                SetLines();
            }
        }

        void OnPostRender()
        {
            if (mainTargetRoot == null || manuallyHandleGizmo) return;

            lineMaterial.SetPass(0);

            Color xColor = (nearAxis == Axis.X) ? (isTransforming) ? selectedColor : hoverColor : this.xColor;
            Color yColor = (nearAxis == Axis.Y) ? (isTransforming) ? selectedColor : hoverColor : this.yColor;
            Color zColor = (nearAxis == Axis.Z) ? (isTransforming) ? selectedColor : hoverColor : this.zColor;
            Color allColor = (nearAxis == Axis.Any) ? (isTransforming) ? selectedColor : hoverColor : this.allColor;

            //Note: The order of drawing the axis decides what gets drawn over what.

            TransformType moveType = TransformType.Move;
            DrawQuads(handleLines.z, GetColor(moveType, this.zColor, zColor, hasTranslatingAxisPlane));
            DrawQuads(handleLines.x, GetColor(moveType, this.xColor, xColor, hasTranslatingAxisPlane));
            DrawQuads(handleLines.y, GetColor(moveType, this.yColor, yColor, hasTranslatingAxisPlane));

            DrawTriangles(handleTriangles.x, GetColor(TransformType.Move, this.xColor, xColor, hasTranslatingAxisPlane));
            DrawTriangles(handleTriangles.y, GetColor(TransformType.Move, this.yColor, yColor, hasTranslatingAxisPlane));
            DrawTriangles(handleTriangles.z, GetColor(TransformType.Move, this.zColor, zColor, hasTranslatingAxisPlane));

            DrawQuads(handlePlanes.z, GetColor(TransformType.Move, this.zColor, zColor, planesOpacity, !hasTranslatingAxisPlane));
            DrawQuads(handlePlanes.x, GetColor(TransformType.Move, this.xColor, xColor, planesOpacity, !hasTranslatingAxisPlane));
            DrawQuads(handlePlanes.y, GetColor(TransformType.Move, this.yColor, yColor, planesOpacity, !hasTranslatingAxisPlane));

            DrawQuads(circlesLines.all, GetColor(TransformType.Rotate, this.allColor, allColor));
            DrawQuads(circlesLines.x, GetColor(TransformType.Rotate, this.xColor, xColor));
            DrawQuads(circlesLines.y, GetColor(TransformType.Rotate, this.yColor, yColor));
            DrawQuads(circlesLines.z, GetColor(TransformType.Rotate, this.zColor, zColor));
        }

        Color GetColor(TransformType type, Color normalColor, Color nearColor, bool forceUseNormal = false)
        {
            return GetColor(type, normalColor, nearColor, false, 1, forceUseNormal);
        }
        Color GetColor(TransformType type, Color normalColor, Color nearColor, float alpha, bool forceUseNormal = false)
        {
            return GetColor(type, normalColor, nearColor, true, alpha, forceUseNormal);
        }
        Color GetColor(TransformType type, Color normalColor, Color nearColor, bool setAlpha, float alpha, bool forceUseNormal = false)
        {
            Color color;
            if (!forceUseNormal && TranslatingTypeContains(type, false))
            {
                color = nearColor;
            }
            else
            {
                color = normalColor;
            }

            if (setAlpha)
            {
                color.a = alpha;
            }

            return color;
        }

        public TransformSpace GetProperTransformSpace()
        {
            return space;
        }

        public bool TransformTypeContains(TransformType type)
        {
            return TransformTypeContains(transformType, type);
        }
        public bool TranslatingTypeContains(TransformType type, bool checkIsTransforming = true)
        {
            TransformType transType = !checkIsTransforming || isTransforming ? translatingType : transformType;
            return TransformTypeContains(transType, type);
        }
        public bool TransformTypeContains(TransformType mainType, TransformType type)
        {
            return ExtTransformType.TransformTypeContains(mainType, type, GetProperTransformSpace());
        }

        public float GetHandleLength(TransformType type, Axis axis = Axis.None, bool multiplyDistanceMultiplier = true)
        {
            float length = handleLength;
            if (multiplyDistanceMultiplier) length *= GetDistanceMultiplier();
            return length;
        }

        void TransformSelected()
        {
            if (mainTargetRoot != null)
            {
                if (nearAxis != Axis.None && Input.GetMouseButtonDown(0))
                {
                    StartCoroutine(TransformSelected(translatingType));
                }
            }
        }

        IEnumerator TransformSelected(TransformType transType)
        {
            isTransforming = true;
            totalRotationAmount = Quaternion.identity;

            Vector3 originalPivot = pivotPoint;

            Vector3 otherAxis1, otherAxis2;
            Vector3 axis = GetNearAxisDirection(out otherAxis1, out otherAxis2);
            Vector3 planeNormal = hasTranslatingAxisPlane ? axis : (transform.position - originalPivot).normalized;
            Vector3 projectedAxis = Vector3.ProjectOnPlane(axis, planeNormal).normalized;
            Vector3 previousMousePosition = Vector3.zero;

            Vector3 currentSnapMovementAmount = Vector3.zero;
            float currentSnapRotationAmount = 0;

            while (!Input.GetMouseButtonUp(0))
            {
                Ray mouseRay = Camera.main.ScreenPointToRay(Input.mousePosition);
                Vector3 mousePosition = Geometry.LinePlaneIntersect(mouseRay.origin, mouseRay.direction, originalPivot, planeNormal);
                bool isSnapping = Input.GetKey(translationSnapping);

                if (previousMousePosition != Vector3.zero && mousePosition != Vector3.zero)
                {
                    if (transType == TransformType.Move)
                    {
                        Vector3 movement = Vector3.zero;

                        if (hasTranslatingAxisPlane)
                        {
                            movement = mousePosition - previousMousePosition;
                        }
                        else
                        {
                            float moveAmount = ExtVector3.MagnitudeInDirection(mousePosition - previousMousePosition, projectedAxis) * moveSpeedMultiplier;
                            movement = axis * moveAmount;
                        }

                        if (isSnapping && movementSnap > 0)
                        {
                            currentSnapMovementAmount += movement;
                            movement = Vector3.zero;

                            if (hasTranslatingAxisPlane)
                            {
                                float amountInAxis1 = ExtVector3.MagnitudeInDirection(currentSnapMovementAmount, otherAxis1);
                                float amountInAxis2 = ExtVector3.MagnitudeInDirection(currentSnapMovementAmount, otherAxis2);

                                float remainder1;
                                float snapAmount1 = CalculateSnapAmount(movementSnap, amountInAxis1, out remainder1);
                                float remainder2;
                                float snapAmount2 = CalculateSnapAmount(movementSnap, amountInAxis2, out remainder2);

                                if (snapAmount1 != 0)
                                {
                                    Vector3 snapMove = (otherAxis1 * snapAmount1);
                                    movement += snapMove;
                                    currentSnapMovementAmount -= snapMove;
                                }
                                if (snapAmount2 != 0)
                                {
                                    Vector3 snapMove = (otherAxis2 * snapAmount2);
                                    movement += snapMove;
                                    currentSnapMovementAmount -= snapMove;
                                }
                            }
                            else
                            {
                                float remainder;
                                float snapAmount = CalculateSnapAmount(movementSnap, currentSnapMovementAmount.magnitude, out remainder);

                                if (snapAmount != 0)
                                {
                                    movement = currentSnapMovementAmount.normalized * snapAmount;
                                    currentSnapMovementAmount = currentSnapMovementAmount.normalized * remainder;
                                }
                            }
                        }

                        for (int i = 0; i < targetRootsOrdered.Count; i++)
                        {
                            Transform target = targetRootsOrdered[i];

                            target.Translate(movement, Space.World);
                        }

                        SetPivotPointOffset(movement);
                    }
                    else if (transType == TransformType.Rotate)
                    {
                        float rotateAmount = 0;
                        Vector3 rotationAxis = axis;

                        if (nearAxis == Axis.Any)
                        {
                            Vector3 rotation = transform.TransformDirection(new Vector3(Input.GetAxis("Mouse Y"), -Input.GetAxis("Mouse X"), 0));
                            Quaternion.Euler(rotation).ToAngleAxis(out rotateAmount, out rotationAxis);
                            rotateAmount *= allRotateSpeedMultiplier;
                        }
                        else
                        {
                            if (circularRotationMethod)
                            {
                                float angle = Vector3.SignedAngle(previousMousePosition - originalPivot, mousePosition - originalPivot, axis);
                                rotateAmount = angle * rotateSpeedMultiplier;
                            }
                            else
                            {
                                Vector3 projected = (nearAxis == Axis.Any || ExtVector3.IsParallel(axis, planeNormal)) ? planeNormal : Vector3.Cross(axis, planeNormal);
                                rotateAmount = (ExtVector3.MagnitudeInDirection(mousePosition - previousMousePosition, projected) * (rotateSpeedMultiplier * 100f)) / GetDistanceMultiplier();
                            }
                        }

                        if (isSnapping && rotationSnap > 0)
                        {
                            currentSnapRotationAmount += rotateAmount;
                            rotateAmount = 0;

                            float remainder;
                            float snapAmount = CalculateSnapAmount(rotationSnap, currentSnapRotationAmount, out remainder);

                            if (snapAmount != 0)
                            {
                                rotateAmount = snapAmount;
                                currentSnapRotationAmount = remainder;
                            }
                        }

                        for (int i = 0; i < targetRootsOrdered.Count; i++)
                        {
                            Transform target = targetRootsOrdered[i];

                            if (pivot == TransformPivot.Pivot)
                            {
                                target.Rotate(rotationAxis, rotateAmount, Space.World);
                            }
                            else if (pivot == TransformPivot.Center)
                            {
                                target.RotateAround(originalPivot, rotationAxis, rotateAmount);
                            }
                        }

                        totalRotationAmount *= Quaternion.Euler(rotationAxis * rotateAmount);
                    }
                }

                previousMousePosition = mousePosition;

                yield return null;
            }

            totalRotationAmount = Quaternion.identity;
            isTransforming = false;
            SetTranslatingAxis(transformType, Axis.None);

            SetPivotPoint();
        }

        float CalculateSnapAmount(float snapValue, float currentAmount, out float remainder)
        {
            remainder = 0;
            if (snapValue <= 0) return currentAmount;

            float currentAmountAbs = Mathf.Abs(currentAmount);
            if (currentAmountAbs > snapValue)
            {
                remainder = currentAmountAbs % snapValue;
                return snapValue * (Mathf.Sign(currentAmount) * Mathf.Floor(currentAmountAbs / snapValue));
            }

            return 0;
        }

        Vector3 GetNearAxisDirection(out Vector3 otherAxis1, out Vector3 otherAxis2)
        {
            otherAxis1 = otherAxis2 = Vector3.zero;

            if (nearAxis != Axis.None)
            {
                if (nearAxis == Axis.X)
                {
                    otherAxis1 = axisInfo.yDirection;
                    otherAxis2 = axisInfo.zDirection;
                    return axisInfo.xDirection;
                }
                if (nearAxis == Axis.Y)
                {
                    otherAxis1 = axisInfo.xDirection;
                    otherAxis2 = axisInfo.zDirection;
                    return axisInfo.yDirection;
                }
                if (nearAxis == Axis.Z)
                {
                    otherAxis1 = axisInfo.xDirection;
                    otherAxis2 = axisInfo.yDirection;
                    return axisInfo.zDirection;
                }
                if (nearAxis == Axis.Any)
                {
                    return Vector3.one;
                }
            }

            return Vector3.zero;
        }

        public void AddTarget(Transform target)
        {
            if (target != null)
            {
                if (targetRoots.ContainsKey(target)) return;
                if (children.Contains(target)) return;

                AddTargetRoot(target);

                SetPivotPoint();
            }
        }

        public void RemoveTarget(Transform target)
        {
            if (target != null)
            {
                if (!targetRoots.ContainsKey(target)) return;

                RemoveTargetRoot(target);

                SetPivotPoint();
            }
        }

        public void ClearTargets()
        {
            targetRoots.Clear();
            targetRootsOrdered.Clear();
            children.Clear();
        }

        public void ClearAndAddTarget(Transform target)
        {
            ClearTargets();
            AddTarget(target);
        }

        void AddTargetRoot(Transform targetRoot)
        {
            targetRoots.Add(targetRoot, new TargetInfo());
            targetRootsOrdered.Add(targetRoot);

            AddAllChildren(targetRoot);
        }
        void RemoveTargetRoot(Transform targetRoot)
        {
            if (targetRoots.Remove(targetRoot))
            {
                targetRootsOrdered.Remove(targetRoot);

                RemoveAllChildren(targetRoot);
            }
        }

        void AddAllChildren(Transform target)
        {
            childrenBuffer.Clear();
            target.GetComponentsInChildren<Transform>(true, childrenBuffer);
            childrenBuffer.Remove(target);

            for (int i = 0; i < childrenBuffer.Count; i++)
            {
                Transform child = childrenBuffer[i];
                children.Add(child);
                RemoveTargetRoot(child); //We do this in case we selected child first and then the parent.
            }

            childrenBuffer.Clear();
        }
        void RemoveAllChildren(Transform target)
        {
            childrenBuffer.Clear();
            target.GetComponentsInChildren<Transform>(true, childrenBuffer);
            childrenBuffer.Remove(target);

            for (int i = 0; i < childrenBuffer.Count; i++)
            {
                children.Remove(childrenBuffer[i]);
            }

            childrenBuffer.Clear();
        }

        public void SetPivotPoint()
        {
            if (mainTargetRoot != null)
            {
                if (pivot == TransformPivot.Pivot)
                {
                    pivotPoint = mainTargetRoot.position;
                }
                else if (pivot == TransformPivot.Center)
                {
                    totalCenterPivotPoint = Vector3.zero;

                    Dictionary<Transform, TargetInfo>.Enumerator targetsEnumerator = targetRoots.GetEnumerator(); //We avoid foreach to avoid garbage.
                    while (targetsEnumerator.MoveNext())
                    {
                        Transform target = targetsEnumerator.Current.Key;
                        TargetInfo info = targetsEnumerator.Current.Value;
                        info.centerPivotPoint = target.GetCenter();

                        totalCenterPivotPoint += info.centerPivotPoint;
                    }

                    totalCenterPivotPoint /= targetRoots.Count;


                    pivotPoint = totalCenterPivotPoint;
                }
            }
        }
        void SetPivotPointOffset(Vector3 offset)
        {
            pivotPoint += offset;
            totalCenterPivotPoint += offset;
        }


        IEnumerator ForceUpdatePivotPointAtEndOfFrame()
        {
            while (this.enabled)
            {
                ForceUpdatePivotPointOnChange();
                yield return waitForEndOFFrame;
            }
        }

        void ForceUpdatePivotPointOnChange()
        {
            if (forceUpdatePivotPointOnChange)
            {
                if (mainTargetRoot != null && !isTransforming)
                {
                    bool hasSet = false;
                    Dictionary<Transform, TargetInfo>.Enumerator targets = targetRoots.GetEnumerator();
                    while (targets.MoveNext())
                    {
                        if (!hasSet)
                        {
                            if (targets.Current.Value.previousPosition != Vector3.zero && targets.Current.Key.position != targets.Current.Value.previousPosition)
                            {
                                SetPivotPoint();
                                hasSet = true;
                            }
                        }

                        targets.Current.Value.previousPosition = targets.Current.Key.position;
                    }
                }
            }
        }

        public void SetTranslatingAxis(TransformType type, Axis axis, Axis planeAxis = Axis.None)
        {
            this.translatingType = type;
            this.nearAxis = axis;
            this.planeAxis = planeAxis;
        }

        public AxisInfo GetAxisInfo()
        {
            AxisInfo currentAxisInfo = axisInfo;

            if (isTransforming && GetProperTransformSpace() == TransformSpace.Global && translatingType == TransformType.Rotate)
            {
                currentAxisInfo.xDirection = totalRotationAmount * Vector3.right;
                currentAxisInfo.yDirection = totalRotationAmount * Vector3.up;
                currentAxisInfo.zDirection = totalRotationAmount * Vector3.forward;
            }

            return currentAxisInfo;
        }

        void SetNearAxis()
        {
            if (isTransforming) return;

            SetTranslatingAxis(transformType, Axis.None);

            if (mainTargetRoot == null) return;

            float distanceMultiplier = GetDistanceMultiplier();
            float handleMinSelectedDistanceCheck = (this.minSelectedDistanceCheck + handleWidth) * distanceMultiplier;

            if (nearAxis == Axis.None && (TransformTypeContains(TransformType.Move)))
            {
                if (nearAxis == Axis.None && TransformTypeContains(TransformType.Move))
                {
                    //Important to check the planes first before the handle tip since it makes selecting the planes easier.
                    float planeMinSelectedDistanceCheck = (this.minSelectedDistanceCheck + planeSize) * distanceMultiplier;
                    HandleNearestPlanes(TransformType.Move, handlePlanes, planeMinSelectedDistanceCheck);

                    if (nearAxis != Axis.None)
                    {
                        planeAxis = nearAxis;
                    }
                    else
                    {
                        float tipMinSelectedDistanceCheck = (this.minSelectedDistanceCheck + triangleSize) * distanceMultiplier;
                        HandleNearestLines(TransformType.Move, handleTriangles, tipMinSelectedDistanceCheck);
                    }
                }

                if (nearAxis == Axis.None)
                {
                    HandleNearestLines(transformType, handleLines, handleMinSelectedDistanceCheck);
                }
            }

            if (nearAxis == Axis.None && TransformTypeContains(TransformType.Rotate))
            {
                HandleNearestLines(TransformType.Rotate, circlesLines, handleMinSelectedDistanceCheck);
            }
        }

        void HandleNearestLines(TransformType type, AxisVectors axisVectors, float minSelectedDistanceCheck)
        {
            float xClosestDistance = ClosestDistanceFromMouseToLines(axisVectors.x);
            float yClosestDistance = ClosestDistanceFromMouseToLines(axisVectors.y);
            float zClosestDistance = ClosestDistanceFromMouseToLines(axisVectors.z);
            float allClosestDistance = ClosestDistanceFromMouseToLines(axisVectors.all);

            HandleNearest(type, xClosestDistance, yClosestDistance, zClosestDistance, allClosestDistance, minSelectedDistanceCheck);
        }

        void HandleNearestPlanes(TransformType type, AxisVectors axisVectors, float minSelectedDistanceCheck)
        {
            float xClosestDistance = ClosestDistanceFromMouseToPlanes(axisVectors.x);
            float yClosestDistance = ClosestDistanceFromMouseToPlanes(axisVectors.y);
            float zClosestDistance = ClosestDistanceFromMouseToPlanes(axisVectors.z);
            float allClosestDistance = ClosestDistanceFromMouseToPlanes(axisVectors.all);

            HandleNearest(type, xClosestDistance, yClosestDistance, zClosestDistance, allClosestDistance, minSelectedDistanceCheck);
        }

        void HandleNearest(TransformType type, float xClosestDistance, float yClosestDistance, float zClosestDistance, float allClosestDistance, float minSelectedDistanceCheck)
        {
            if (xClosestDistance <= minSelectedDistanceCheck && xClosestDistance <= yClosestDistance && xClosestDistance <= zClosestDistance) SetTranslatingAxis(type, Axis.X);
            else if (yClosestDistance <= minSelectedDistanceCheck && yClosestDistance <= xClosestDistance && yClosestDistance <= zClosestDistance) SetTranslatingAxis(type, Axis.Y);
            else if (zClosestDistance <= minSelectedDistanceCheck && zClosestDistance <= xClosestDistance && zClosestDistance <= yClosestDistance) SetTranslatingAxis(type, Axis.Z);
            else if (type == TransformType.Rotate && mainTargetRoot != null)
            {
                Ray mouseRay = Camera.main.ScreenPointToRay(Input.mousePosition);
                Vector3 mousePlaneHit = Geometry.LinePlaneIntersect(mouseRay.origin, mouseRay.direction, pivotPoint, (transform.position - pivotPoint).normalized);
                if ((pivotPoint - mousePlaneHit).sqrMagnitude <= (GetHandleLength(TransformType.Rotate)).Squared()) SetTranslatingAxis(type, Axis.Any);
            }
        }

        float ClosestDistanceFromMouseToLines(List<Vector3> lines)
        {
            Ray mouseRay = Camera.main.ScreenPointToRay(Input.mousePosition);

            float closestDistance = float.MaxValue;
            for (int i = 0; i + 1 < lines.Count; i++)
            {
                IntersectPoints points = Geometry.ClosestPointsOnSegmentToLine(lines[i], lines[i + 1], mouseRay.origin, mouseRay.direction);
                float distance = Vector3.Distance(points.first, points.second);
                if (distance < closestDistance)
                {
                    closestDistance = distance;
                }
            }
            return closestDistance;
        }

        float ClosestDistanceFromMouseToPlanes(List<Vector3> planePoints)
        {
            float closestDistance = float.MaxValue;

            if (planePoints.Count >= 4)
            {
                Ray mouseRay = Camera.main.ScreenPointToRay(Input.mousePosition);

                for (int i = 0; i < planePoints.Count; i += 4)
                {
                    Plane plane = new Plane(planePoints[i], planePoints[i + 1], planePoints[i + 2]);

                    float distanceToPlane;
                    if (plane.Raycast(mouseRay, out distanceToPlane))
                    {
                        Vector3 pointOnPlane = mouseRay.origin + (mouseRay.direction * distanceToPlane);
                        Vector3 planeCenter = (planePoints[0] + planePoints[1] + planePoints[2] + planePoints[3]) / 4f;

                        float distance = Vector3.Distance(planeCenter, pointOnPlane);
                        if (distance < closestDistance)
                        {
                            closestDistance = distance;
                        }
                    }
                }
            }

            return closestDistance;
        }

        void SetAxisInfo()
        {
            if (mainTargetRoot != null)
            {
                axisInfo.Set(mainTargetRoot, pivotPoint, GetProperTransformSpace());
            }
        }

        public float GetDistanceMultiplier()
        {
            if (mainTargetRoot == null) return 0f;

            if (Camera.main.orthographic) return Mathf.Max(.01f, Camera.main.orthographicSize * 2f);
            return Mathf.Max(.01f, Mathf.Abs(ExtVector3.MagnitudeInDirection(pivotPoint - transform.position, Camera.main.transform.forward)));
        }

        void SetLines()
        {
            SetHandleLines();
            SetHandlePlanes();
            SetHandleTriangles();
            SetHandleSquares();
            SetCircles(GetAxisInfo(), circlesLines);
        }

        void SetHandleLines()
        {
            handleLines.Clear();

            if (TranslatingTypeContains(TransformType.Move))
            {
                float lineWidth = handleWidth * GetDistanceMultiplier();

                float xLineLength = 0;
                float yLineLength = 0;
                float zLineLength = 0;
                if (TranslatingTypeContains(TransformType.Move))
                {
                    xLineLength = yLineLength = zLineLength = GetHandleLength(TransformType.Move);
                }

                AddQuads(pivotPoint, axisInfo.xDirection, axisInfo.yDirection, axisInfo.zDirection, xLineLength, lineWidth, handleLines.x);
                AddQuads(pivotPoint, axisInfo.yDirection, axisInfo.xDirection, axisInfo.zDirection, yLineLength, lineWidth, handleLines.y);
                AddQuads(pivotPoint, axisInfo.zDirection, axisInfo.xDirection, axisInfo.yDirection, zLineLength, lineWidth, handleLines.z);
            }
        }
        int AxisDirectionMultiplier(Vector3 direction, Vector3 otherDirection)
        {
            return ExtVector3.IsInDirection(direction, otherDirection) ? 1 : -1;
        }

        void SetHandlePlanes()
        {
            handlePlanes.Clear();

            if (TranslatingTypeContains(TransformType.Move))
            {
                Vector3 pivotToCamera = Camera.main.transform.position - pivotPoint;
                float cameraXSign = Mathf.Sign(Vector3.Dot(axisInfo.xDirection, pivotToCamera));
                float cameraYSign = Mathf.Sign(Vector3.Dot(axisInfo.yDirection, pivotToCamera));
                float cameraZSign = Mathf.Sign(Vector3.Dot(axisInfo.zDirection, pivotToCamera));

                float planeSize = this.planeSize;
                planeSize *= GetDistanceMultiplier();

                Vector3 xDirection = (axisInfo.xDirection * planeSize) * cameraXSign;
                Vector3 yDirection = (axisInfo.yDirection * planeSize) * cameraYSign;
                Vector3 zDirection = (axisInfo.zDirection * planeSize) * cameraZSign;

                Vector3 xPlaneCenter = pivotPoint + (yDirection + zDirection);
                Vector3 yPlaneCenter = pivotPoint + (xDirection + zDirection);
                Vector3 zPlaneCenter = pivotPoint + (xDirection + yDirection);

                AddQuad(xPlaneCenter, axisInfo.yDirection, axisInfo.zDirection, planeSize, handlePlanes.x);
                AddQuad(yPlaneCenter, axisInfo.xDirection, axisInfo.zDirection, planeSize, handlePlanes.y);
                AddQuad(zPlaneCenter, axisInfo.xDirection, axisInfo.yDirection, planeSize, handlePlanes.z);
            }
        }

        void SetHandleTriangles()
        {
            handleTriangles.Clear();

            if (TranslatingTypeContains(TransformType.Move))
            {
                float triangleLength = triangleSize * GetDistanceMultiplier();
                AddTriangles(axisInfo.GetXAxisEnd(GetHandleLength(TransformType.Move)), axisInfo.xDirection, axisInfo.yDirection, axisInfo.zDirection, triangleLength, handleTriangles.x);
                AddTriangles(axisInfo.GetYAxisEnd(GetHandleLength(TransformType.Move)), axisInfo.yDirection, axisInfo.xDirection, axisInfo.zDirection, triangleLength, handleTriangles.y);
                AddTriangles(axisInfo.GetZAxisEnd(GetHandleLength(TransformType.Move)), axisInfo.zDirection, axisInfo.yDirection, axisInfo.xDirection, triangleLength, handleTriangles.z);
            }
        }

        void AddTriangles(Vector3 axisEnd, Vector3 axisDirection, Vector3 axisOtherDirection1, Vector3 axisOtherDirection2, float size, List<Vector3> resultsBuffer)
        {
            Vector3 endPoint = axisEnd + (axisDirection * (size * 2f));
            Square baseSquare = GetBaseSquare(axisEnd, axisOtherDirection1, axisOtherDirection2, size / 2f);

            resultsBuffer.Add(baseSquare.bottomLeft);
            resultsBuffer.Add(baseSquare.topLeft);
            resultsBuffer.Add(baseSquare.topRight);
            resultsBuffer.Add(baseSquare.topLeft);
            resultsBuffer.Add(baseSquare.bottomRight);
            resultsBuffer.Add(baseSquare.topRight);

            for (int i = 0; i < 4; i++)
            {
                resultsBuffer.Add(baseSquare[i]);
                resultsBuffer.Add(baseSquare[i + 1]);
                resultsBuffer.Add(endPoint);
            }
        }

        void SetHandleSquares()
        {
            handleSquares.Clear();
        }

        void AddSquares(Vector3 axisStart, Vector3 axisDirection, Vector3 axisOtherDirection1, Vector3 axisOtherDirection2, float size, List<Vector3> resultsBuffer)
        {
            AddQuads(axisStart, axisDirection, axisOtherDirection1, axisOtherDirection2, size, size * .5f, resultsBuffer);
        }
        void AddQuads(Vector3 axisStart, Vector3 axisDirection, Vector3 axisOtherDirection1, Vector3 axisOtherDirection2, float length, float width, List<Vector3> resultsBuffer)
        {
            Vector3 axisEnd = axisStart + (axisDirection * length);
            AddQuads(axisStart, axisEnd, axisOtherDirection1, axisOtherDirection2, width, resultsBuffer);
        }
        void AddQuads(Vector3 axisStart, Vector3 axisEnd, Vector3 axisOtherDirection1, Vector3 axisOtherDirection2, float width, List<Vector3> resultsBuffer)
        {
            Square baseRectangle = GetBaseSquare(axisStart, axisOtherDirection1, axisOtherDirection2, width);
            Square baseRectangleEnd = GetBaseSquare(axisEnd, axisOtherDirection1, axisOtherDirection2, width);

            resultsBuffer.Add(baseRectangle.bottomLeft);
            resultsBuffer.Add(baseRectangle.topLeft);
            resultsBuffer.Add(baseRectangle.topRight);
            resultsBuffer.Add(baseRectangle.bottomRight);

            resultsBuffer.Add(baseRectangleEnd.bottomLeft);
            resultsBuffer.Add(baseRectangleEnd.topLeft);
            resultsBuffer.Add(baseRectangleEnd.topRight);
            resultsBuffer.Add(baseRectangleEnd.bottomRight);

            for (int i = 0; i < 4; i++)
            {
                resultsBuffer.Add(baseRectangle[i]);
                resultsBuffer.Add(baseRectangleEnd[i]);
                resultsBuffer.Add(baseRectangleEnd[i + 1]);
                resultsBuffer.Add(baseRectangle[i + 1]);
            }
        }

        void AddQuad(Vector3 axisStart, Vector3 axisOtherDirection1, Vector3 axisOtherDirection2, float width, List<Vector3> resultsBuffer)
        {
            Square baseRectangle = GetBaseSquare(axisStart, axisOtherDirection1, axisOtherDirection2, width);

            resultsBuffer.Add(baseRectangle.bottomLeft);
            resultsBuffer.Add(baseRectangle.topLeft);
            resultsBuffer.Add(baseRectangle.topRight);
            resultsBuffer.Add(baseRectangle.bottomRight);
        }

        Square GetBaseSquare(Vector3 axisEnd, Vector3 axisOtherDirection1, Vector3 axisOtherDirection2, float size)
        {
            Square square;
            Vector3 offsetUp = ((axisOtherDirection1 * size) + (axisOtherDirection2 * size));
            Vector3 offsetDown = ((axisOtherDirection1 * size) - (axisOtherDirection2 * size));
            //These might not really be the proper directions, as in the bottomLeft might not really be at the bottom left...
            square.bottomLeft = axisEnd + offsetDown;
            square.topLeft = axisEnd + offsetUp;
            square.bottomRight = axisEnd - offsetUp;
            square.topRight = axisEnd - offsetDown;
            return square;
        }

        void SetCircles(AxisInfo axisInfo, AxisVectors axisVectors)
        {
            axisVectors.Clear();

            if (TranslatingTypeContains(TransformType.Rotate))
            {
                float circleLength = GetHandleLength(TransformType.Rotate);
                AddCircle(pivotPoint, axisInfo.xDirection, circleLength, axisVectors.x);
                AddCircle(pivotPoint, axisInfo.yDirection, circleLength, axisVectors.y);
                AddCircle(pivotPoint, axisInfo.zDirection, circleLength, axisVectors.z);
                AddCircle(pivotPoint, (pivotPoint - transform.position).normalized, circleLength, axisVectors.all, false);
            }
        }

        void AddCircle(Vector3 origin, Vector3 axisDirection, float size, List<Vector3> resultsBuffer, bool depthTest = true)
        {
            Vector3 up = axisDirection.normalized * size;
            Vector3 forward = Vector3.Slerp(up, -up, .5f);
            Vector3 right = Vector3.Cross(up, forward).normalized * size;

            Matrix4x4 matrix = new Matrix4x4();

            matrix[0] = right.x;
            matrix[1] = right.y;
            matrix[2] = right.z;

            matrix[4] = up.x;
            matrix[5] = up.y;
            matrix[6] = up.z;

            matrix[8] = forward.x;
            matrix[9] = forward.y;
            matrix[10] = forward.z;

            Vector3 lastPoint = origin + matrix.MultiplyPoint3x4(new Vector3(Mathf.Cos(0), 0, Mathf.Sin(0)));
            Vector3 nextPoint = Vector3.zero;
            float multiplier = 360f / circleDetail;

            Plane plane = new Plane((transform.position - pivotPoint).normalized, pivotPoint);

            float circleHandleWidth = handleWidth * GetDistanceMultiplier();

            for (int i = 0; i < circleDetail + 1; i++)
            {
                nextPoint.x = Mathf.Cos((i * multiplier) * Mathf.Deg2Rad);
                nextPoint.z = Mathf.Sin((i * multiplier) * Mathf.Deg2Rad);
                nextPoint.y = 0;

                nextPoint = origin + matrix.MultiplyPoint3x4(nextPoint);

                if (!depthTest || plane.GetSide(lastPoint))
                {
                    Vector3 centerPoint = (lastPoint + nextPoint) * .5f;
                    Vector3 upDirection = (centerPoint - origin).normalized;
                    AddQuads(lastPoint, nextPoint, upDirection, axisDirection, circleHandleWidth, resultsBuffer);
                }

                lastPoint = nextPoint;
            }
        }

        void DrawLines(List<Vector3> lines, Color color)
        {
            if (lines.Count == 0) return;

            GL.Begin(GL.LINES);
            GL.Color(color);

            for (int i = 0; i < lines.Count; i += 2)
            {
                GL.Vertex(lines[i]);
                GL.Vertex(lines[i + 1]);
            }

            GL.End();
        }

        void DrawTriangles(List<Vector3> lines, Color color)
        {
            if (lines.Count == 0) return;

            GL.Begin(GL.TRIANGLES);
            GL.Color(color);

            for (int i = 0; i < lines.Count; i += 3)
            {
                GL.Vertex(lines[i]);
                GL.Vertex(lines[i + 1]);
                GL.Vertex(lines[i + 2]);
            }

            GL.End();
        }

        void DrawQuads(List<Vector3> lines, Color color)
        {
            if (lines.Count == 0) return;

            GL.Begin(GL.QUADS);
            GL.Color(color);

            for (int i = 0; i < lines.Count; i += 4)
            {
                GL.Vertex(lines[i]);
                GL.Vertex(lines[i + 1]);
                GL.Vertex(lines[i + 2]);
                GL.Vertex(lines[i + 3]);
            }

            GL.End();
        }

        void DrawFilledCircle(List<Vector3> lines, Color color)
        {
            if (lines.Count == 0) return;

            Vector3 center = Vector3.zero;
            for (int i = 0; i < lines.Count; i++)
            {
                center += lines[i];
            }
            center /= lines.Count;

            GL.Begin(GL.TRIANGLES);
            GL.Color(color);

            for (int i = 0; i + 1 < lines.Count; i++)
            {
                GL.Vertex(lines[i]);
                GL.Vertex(lines[i + 1]);
                GL.Vertex(center);
            }

            GL.End();
        }

        void SetMaterial()
        {
            if (lineMaterial == null)
            {
                lineMaterial = new Material(WorldEditor.LineShader);
            }
        }
    }
    public struct Square
    {
        public Vector3 bottomLeft;
        public Vector3 bottomRight;
        public Vector3 topLeft;
        public Vector3 topRight;

        public Vector3 this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0:
                        return this.bottomLeft;
                    case 1:
                        return this.topLeft;
                    case 2:
                        return this.topRight;
                    case 3:
                        return this.bottomRight;
                    case 4:
                        return this.bottomLeft; //so we wrap around back to start
                    default:
                        return Vector3.zero;
                }
            }
        }
    }
    public class TargetInfo
    {
        public Vector3 centerPivotPoint;

        public Vector3 previousPosition;
    }
    public struct IntersectPoints
    {
        public Vector3 first;
        public Vector3 second;

        public IntersectPoints(Vector3 first, Vector3 second)
        {
            this.first = first;
            this.second = second;
        }
    }
    public enum TransformSpace { Global, Local }
    public enum TransformType { Move, Rotate }
    public enum TransformPivot { Pivot, Center }
    public enum Axis { None, X, Y, Z, Any }
    public class AxisVectors
    {
        public List<Vector3> x = new List<Vector3>();
        public List<Vector3> y = new List<Vector3>();
        public List<Vector3> z = new List<Vector3>();
        public List<Vector3> all = new List<Vector3>();

        public void Add(AxisVectors axisVectors)
        {
            x.AddRange(axisVectors.x);
            y.AddRange(axisVectors.y);
            z.AddRange(axisVectors.z);
            all.AddRange(axisVectors.all);
        }

        public void Clear()
        {
            x.Clear();
            y.Clear();
            z.Clear();
            all.Clear();
        }
    }
    public struct AxisInfo
    {
        public Vector3 pivot;
        public Vector3 xDirection;
        public Vector3 yDirection;
        public Vector3 zDirection;

        public void Set(Transform target, Vector3 pivot, TransformSpace space)
        {
            if (space == TransformSpace.Global)
            {
                xDirection = Vector3.right;
                yDirection = Vector3.up;
                zDirection = Vector3.forward;
            }
            else if (space == TransformSpace.Local)
            {
                xDirection = target.right;
                yDirection = target.up;
                zDirection = target.forward;
            }

            this.pivot = pivot;
        }

        public Vector3 GetXAxisEnd(float size)
        {
            return pivot + (xDirection * size);
        }
        public Vector3 GetYAxisEnd(float size)
        {
            return pivot + (yDirection * size);
        }
        public Vector3 GetZAxisEnd(float size)
        {
            return pivot + (zDirection * size);
        }
        public Vector3 GetAxisEnd(Vector3 direction, float size)
        {
            return pivot + (direction * size);
        }
    }
    public static class ExtMathf
    {
        public static float Squared(this float value)
        {
            return value * value;
        }

        public static float SafeDivide(float value, float divider)
        {
            if (divider == 0) return 0;
            return value / divider;
        }
    }
    public static class ExtTransform
    {
        public static Vector3 GetCenter(this Transform transform)
        {
            Bounds totalBounds = new Bounds(transform.position, Vector3.zero);
            GetCenterAll(transform, ref totalBounds);
            return totalBounds.center;
        }
        static void GetCenterAll(this Transform transform, ref Bounds currentTotalBounds)
        {
            Renderer renderer = transform.GetComponent<Renderer>();
            if (renderer != null)
            {
                currentTotalBounds.Encapsulate(renderer.bounds);
            }
            else
            {
                currentTotalBounds.Encapsulate(transform.position);
            }

            for (int i = 0; i < transform.childCount; i++)
            {
                transform.GetChild(i).GetCenterAll(ref currentTotalBounds);
            }
        }
    }
    public static class ExtTransformType
    {
        public static bool TransformTypeContains(this TransformType mainType, TransformType type, TransformSpace space)
        {
            if (type == mainType) return true;
            return false;
        }
    }
    public static class ExtVector3
    {
        public static float MagnitudeInDirection(Vector3 vector, Vector3 direction, bool normalizeParameters = true)
        {
            if (normalizeParameters) direction.Normalize();
            return Vector3.Dot(vector, direction);
        }

        public static Vector3 Abs(this Vector3 vector)
        {
            return new Vector3(Mathf.Abs(vector.x), Mathf.Abs(vector.y), Mathf.Abs(vector.z));
        }

        public static bool IsParallel(Vector3 direction, Vector3 otherDirection, float precision = .0001f)
        {
            return Vector3.Cross(direction, otherDirection).sqrMagnitude < precision;
        }

        public static bool IsInDirection(Vector3 direction, Vector3 otherDirection)
        {
            return Vector3.Dot(direction, otherDirection) > 0f;
        }
    }
    public static class Geometry
    {
        public static float LinePlaneDistance(Vector3 linePoint, Vector3 lineVec, Vector3 planePoint, Vector3 planeNormal)
        {
            //calculate the distance between the linePoint and the line-plane intersection point
            float dotNumerator = Vector3.Dot((planePoint - linePoint), planeNormal);
            float dotDenominator = Vector3.Dot(lineVec, planeNormal);

            //line and plane are not parallel
            if (dotDenominator != 0f)
            {
                return dotNumerator / dotDenominator;
            }

            return 0;
        }

        //Note that the line is infinite, this is not a line-segment plane intersect
        public static Vector3 LinePlaneIntersect(Vector3 linePoint, Vector3 lineVec, Vector3 planePoint, Vector3 planeNormal)
        {
            float distance = LinePlaneDistance(linePoint, lineVec, planePoint, planeNormal);

            //line and plane are not parallel
            if (distance != 0f)
            {
                return linePoint + (lineVec * distance);
            }

            return Vector3.zero;
        }

        //Returns 2 points since on line 1 there will be a closest point to line 2, and on line 2 there will be a closest point to line 1.
        public static IntersectPoints ClosestPointsOnTwoLines(Vector3 point1, Vector3 point1Direction, Vector3 point2, Vector3 point2Direction)
        {
            IntersectPoints intersections = new IntersectPoints();

            //I dont think we need to normalize
            //point1Direction.Normalize();
            //point2Direction.Normalize();

            float a = Vector3.Dot(point1Direction, point1Direction);
            float b = Vector3.Dot(point1Direction, point2Direction);
            float e = Vector3.Dot(point2Direction, point2Direction);

            float d = a * e - b * b;

            //This is a check if parallel, howeverm since we are not normalizing the directions, it seems even if they are parallel they will not == 0
            //so they will get past this point, but its seems to be alright since it seems to still give a correct point (although a point very fary away).
            //Also, if they are parallel and we dont normalize, the deciding point seems randomly choses on the lines, which while is still correct,
            //our ClosestPointsOnTwoLineSegments gets undesireable results when on corners. (for example when using it in our ClosestPointOnTriangleToLine).
            if (d != 0f)
            {
                Vector3 r = point1 - point2;
                float c = Vector3.Dot(point1Direction, r);
                float f = Vector3.Dot(point2Direction, r);

                float s = (b * f - c * e) / d;
                float t = (a * f - c * b) / d;

                intersections.first = point1 + point1Direction * s;
                intersections.second = point2 + point2Direction * t;
            }
            else
            {
                //Lines are parallel, select any points next to eachother
                intersections.first = point1;
                intersections.second = point2 + Vector3.Project(point1 - point2, point2Direction);
            }

            return intersections;
        }

        public static IntersectPoints ClosestPointsOnSegmentToLine(Vector3 segment0, Vector3 segment1, Vector3 linePoint, Vector3 lineDirection)
        {
            IntersectPoints closests = ClosestPointsOnTwoLines(segment0, segment1 - segment0, linePoint, lineDirection);
            closests.first = ClampToSegment(closests.first, segment0, segment1);

            return closests;
        }

        //Assumes the point is already on the line somewhere
        public static Vector3 ClampToSegment(Vector3 point, Vector3 linePoint1, Vector3 linePoint2)
        {
            Vector3 lineDirection = linePoint2 - linePoint1;

            if (!ExtVector3.IsInDirection(point - linePoint1, lineDirection))
            {
                point = linePoint1;
            }
            else if (ExtVector3.IsInDirection(point - linePoint2, lineDirection))
            {
                point = linePoint2;
            }

            return point;
        }
    }
}